import fasttext

from noisemix import noise, formats
from noisemix import utils

NOISEMIX_SUFFIX = '.nmx'


def mix(dataset, versions=1, format=None):
    """Generates a noisy version of a dataset"""

    if format:
        format = eval('formats.' + args.format)  # string to cls
        format = format()  # cls to instance

    for line in dataset:

        # print(line)

        if format:
            line, data = format.before(line)

        for i in range(0, versions):
            line = mix_sentence(line);
            if format:
                line = format.after(line, data)
            yield line


def mix_sentence(sentence):
    utils.rand_init()
    words = []
    for word in sentence.split(' '):
        words.append(noise.randnoise(word))
    # noisemix.append_noise(words)
    sentence = (' ').join(words)
    return noise.sentence_noise(sentence)


def _read(path):
    with open(path, 'r', encoding='utf-8') as f:
        for line in f:
            yield line


def _write(dataset, path):
    with open(path, 'w', encoding='utf-8') as f:
        for line in dataset:
            f.write(line)


def _print(dataset, max_lines=None):
    line_count = 0
    for line in dataset:
        print(line)
        line_count += 1
        if line_count == max_lines:
            return


if __name__ == '__main__':

    import argparse

    parser = argparse.ArgumentParser(description='Generate a noisy version of dataset')
    parser.add_argument('path', metavar='path', type=str,
                        help='the paths to the input data file')
    parser.add_argument('-format', type=str, choices=[None, 'fastText'],
                        help='the format of the input data file', required=True)
    parser.add_argument('-versions', type=int, default=1,
                        help='how many versions to generate per line')
    parser.add_argument('--print', type=int, nargs='?', const=max,
                        help='print the output instead of writing to file, optional value for max lines')
    args = parser.parse_args()

    dataset = _read(args.path)

    mixed_dataset = mix(dataset, versions=args.versions, format=format)

    if args.print:
        _print(mixed_dataset, args.print)
    else:
        _write(mixed_dataset, args.path + NOISEMIX_SUFFIX)



import string

import random
from datetime import datetime
random.seed(datetime.now())
from noisemix import utils

P_MAX = 3
P_FREQ = 0.2
MIN_LEN = 3
SENTENCE_MIN_LEN = 1
from noisemix.perturbation import Perturbation
import noisemix.utils
LANGUAGE = 'en'
KEYBOARD = 'qwerty'
NOISE_TYPES = []
SENTENCE_NOISE_TYPES = []


# add_letter,repeat_letter,remove_letter,lowercase,remove_punct,word_swap,char_swap

def default_types():
    result = []
    result.append(Perturbation(name="add_letter", function=add_letter))
    result.append(Perturbation(name="repeat_letter", function=repeat_letter))
    result.append(Perturbation(name="remove_letter", function=remove_letter))
    result.append(Perturbation(name="lowercase", function=lowercase))
    result.append(Perturbation(name="remove_punct", function=remove_punct))
    result.append(Perturbation(name="word_swap", function=word_swap))
    result.append(Perturbation(name="char_swap", function=char_swap))
    result.append(Perturbation(name="flip_letters", function=flip_letters))
    result.append(Perturbation(name="typo_qwerty", function=typo_qwerty))
    return result

def sentence_types():
    result = []
    result.append(Perturbation(name="remove_space", function=remove_space, repeat_num=1))
    result.append(Perturbation(name="flip_words", function=flip_words, repeat_num=1))
    return result

def sentence_noise(sentence, p_max=P_MAX):
    words= sentence.split(' ')
    for p in random.sample(SENTENCE_NOISE_TYPES, SENTENCE_MIN_LEN):
        repeat_num = p.repeat_num
        while True:
            if repeat_num == 0:
                break
            if random.random() < p.frequency:
                if len(words) >= SENTENCE_MIN_LEN:
                    words = p.function(words)
            repeat_num = repeat_num - 1
    sentence = (' ').join(words)

    return sentence


def remove_space(words):
    pos= random.randint(0, len(words) - 1)
    words[pos:pos+2] = [''.join(words[pos:pos+2])]
    return words

def flip_words(words):
    result=random.sample(words, 2)
    a, b = words.index(result[0]), words.index(result[1])
    words[b], words[a] = words[a], words[b]
    return words


def randnoise(s, p_max=P_MAX):
    """ Apply random noisification and return the new string """

    # The distribution is designed so that there will be fewer sentences with
    # more noisemix types.

    for p in random.sample(NOISE_TYPES, p_max):
        repeat_num = p.repeat_num
        while True:
            if repeat_num == 0:
                break
            if random.random() < p.frequency:
                if len(s) >= MIN_LEN:
                    s = p.function(s)
            repeat_num = repeat_num - 1
    return s


def _rand_pos(s):
    return utils.rand_index(s)


def flip_letters(s):
    i, j = _rand_pos(s), _rand_pos(s)
    # s[b], s[a] = s[a], s[b]
    if i > j:
        i, j = j, i  # swaps i and j
    s=s[:i]+s[j] + s[i + 1:j] + s[i] + s[j + 1:]
    return s

def repeat_letter(s):
    # print("addition change---before:", s)
    pos = _rand_pos(s)
    s = s[:pos] + s[pos] + s[pos:]
    # print("addition change---after:", s)
    return s


def remove_letter(s):
    # print("subtraction change---before:", word)
    pos = _rand_pos(s)
    s = s[:pos] + s[(pos + 1):]
    # print("subtraction change---after:", word)
    return s


def lowercase(s):
    # print("case change---before:", s)
    s = s.lower()
    # print("case change---after:", s)
    return s


punct_trans = str.maketrans('', '', string.punctuation)


def remove_punct(s):
    return s.translate(punct_trans)


from noisemix.data import word_transforms, char_transforms, keyboard_transforms, letter_transforms

def add_letter(s):
    # print("addition change---before:", s)
    pos = _rand_pos(s)
    addition = random.choice(letter_transforms[LANGUAGE])
    s = s[:pos] + addition + s[pos:]
    # print("addition change---after:", s)
    return s


def word_swap(s):
    for t in word_transforms[LANGUAGE]:
        if s in t:
            return random.sample(t, 1)[0]
    return s


def char_swap(s):
    for t in char_transforms[LANGUAGE]:
        for c in s:
            if c in t:
                return s.replace(c, random.sample(t, 1)[0])
    return s

def typo_qwerty(s):
   # https://stackoverflow.com/a/7621105/4486860
   pos = _rand_pos(s)
   if(s[pos] in keyboard_transforms[KEYBOARD]):
       neighborList = keyboard_transforms[KEYBOARD][s[pos]]
   else:
       return s
   tempList=[]
   for key, value in neighborList:
        if value <= 2:
            tempList.append(key)
        else:
            break
   value= random.sample(tempList, 1)[0]
   s = s[:pos] + value + s[pos:]
   return s


NOISE_TYPES = default_types()
SENTENCE_NOISE_TYPES=sentence_types()
# print("Noise types:", *map(lambda p: p.__name__, PERTURBATIONS))




class Format:

    def skip(self, line):
        False

    def before(self, line):
        return s, None

    def after(self, line, data):
        return s

class fastText(Format):

    LABEL_PREFIX = '__label__'

    def before(self, line):
        labels = []
        _ = []
        for w in line.split(' '):
            if w.startswith(fastText.LABEL_PREFIX):
                labels.append(w)
            else:
                _.append(w)
        return ' '.join(_), labels

    def after(self, line, data):
        return ' '.join(data) + ' ' + line



class Perturbation:
    name = "default"
    function=None

    def __init__(self, name, function,frequency=0.2,repeat_num=1):
        self.name = name
        self.function = function
        self.frequency = frequency
        self.repeat_num = repeat_num

    def change_frequency(self, frequency):
        self.frequency = frequency

    def change_repeatable(self, repeatable):
        self.isRepeatable = repeatable


from setuptools import setup

setup(name='noisemix',
      version='0.1',
      description='NoiseMix is a library for data generation for text datasets.',
      url='https://gitlab.com/hetazotutyun/NoiseMix',
      author='Anooshik Vartnian',
      author_email='anooshik_vartanian@edu.aua.com',
      license='MIT',
      packages=['noisemix'],
      zip_safe=False)


# After randomly initing those lists with 100 random numbers in the correct range
from random import choices, random

RAND_INDICES={}
INDEX_LENGTH=100
LONGEST_WORD_LENGTH=27
POS = {};
def rand_init(val=0):
    if(val==0):
        for i in range(2, LONGEST_WORD_LENGTH):
            RAND_INDICES[i]=choices(range(0,i), k=INDEX_LENGTH)
            POS[i]=0
    else:
        RAND_INDICES[val] = choices(range(0, val), k=INDEX_LENGTH)
        POS[val] = 0
    return

def rand_index(word):
    length=len(word)
    if length<=1:
        return 0
    # Get the current position in the list of random indices for words of length n
    if length in POS:
        pos = POS[length]
    else:
        rand_init(length)
        pos = POS[length]
    # Get the index from that position in the list
    index = RAND_INDICES[length][pos]

    # Update the position in the list of random indices
    if pos + 1 == INDEX_LENGTH:
        POS[length] = random(0,INDEX_LENGTH) # Another option would be to init this with a random number between 0 and 99.
    else:
        POS[length] = pos + 1
    return index


import operator
from math import sqrt, ceil

def transform_keys (keyboard):
    temp={}
    for row,rowS in zip(keyboard["uppercase"], keyboard["lowercase"]):
        y1 =keyboard["uppercase"].index(row)
        for letter, letterS in zip(row, rowS):
            if letter == '':
                continue
            x1=row.index(letter)
            temp[letter]={}
            temp[letterS]={}
            for row2, row2S in zip(keyboard["uppercase"],keyboard["lowercase"]):
                y2= keyboard["uppercase"].index(row2)
                for letter2, letter2S in zip(row2, row2S):
                    if letter2 == '' or letter==letter2:
                        continue
                    x2 = row2.index(letter2)
                    dist=ceil(sqrt(pow((y1 - y2), 2) + pow((x1 - x2), 2)))
                    temp[letterS][letter2S]=dist
                    temp[letter][letter2] =dist
            temp[letterS] = sorted(temp[letterS].items(), key=operator.itemgetter(1))
            temp[letter]=sorted(temp[letter].items(), key=operator.itemgetter(1))
    return temp